unit janTabBar;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  CommCtrl, ComCtrls, StdCtrls, ExtCtrls, Math, DsgnIntf, imgList;

type
  // -- TjanTabBar ----------------------------------------------
  TjanTabBar = class;

  // -- TjanTabSection -------------------------------------------------

  TjanTabSection = class( TScrollBox)
    private    // -- Private Declarations -------
      FActive                : Boolean;
      FOnHide                : TNotifyEvent;
      FOnShow                : TNotifyEvent;
      FBackBitmap: TBitmap;
      FHighLightColor: TColor;
      FTabColor: TColor;
      FImageIndex: TImageIndex;
      procedure BackBitmapChanged(Sender: TObject);
      function GetTabIndex : Integer;
      procedure SetActive( Value: Boolean );
      procedure SetTabIndex( Value: Integer );
      procedure SetBackBitmap(const Value: TBitmap);
      procedure SetHighLightColor(const Value: TColor);
      procedure SetTabColor(const Value: TColor);
      procedure SetImageIndex(Value: TImageIndex);
    protected  // -- Protected Declarations -----
      procedure CMTextChanged( var Msg: TMessage ); message CM_TEXTCHANGED;
      procedure CMFontChanged( var Msg: TMessage ); message CM_FONTCHANGED;
      function ParentAsZC : TjanTabBar;
      function ParentIsZC : Boolean;


    public     // -- Public Declarations --------
      {TjanTabSection is a seperate section within TjanTabBar, an Outlook style bar with scrollable sections that can be opened and closed by clicking the section tab.}
      {}
      {Section Tabs can have a tiled background image and have individual font and hottrack color.}
      {}
      {Sections are added and moved at design time with the component editor.}
      constructor Create( AOwner: TComponent ); override;
      {Creates an instance of TjanTabSection.}
      destructor Destroy; override;
      {Destroys an instance of TjanTabSection.}
      property Active          : Boolean            read FActive         write SetActive;
      {Determines if the instance is the active section within a TjanTabBar instance.}
    published  // -- Published Declarations -----
      property BackBitmap: TBitmap read FBackBitmap write SetBackBitmap;
      {Determines the tiled background of the section tab}
      property Caption;
      {Specifies a text string that identifies the control to the user.}
      property Color;
      {Determines the background color of the Section.}
      property TabColor : TColor read FTabColor write SetTabColor;
      {Determines the color of the section tabs.}
      property ImageIndex: TImageIndex read FImageIndex write SetImageIndex default -1;
      {Determines which image appears on the section tab.}
      property HighLightColor: TColor read FHighLightColor write SetHighLightColor;
      {Determines the hottrack caption highlight color.}
      property Enabled;
      {Controls whether the control responds to mouse, keyboard, and timer events. }
      property TabIndex        : Integer            read GetTabIndex     write SetTabIndex;
      {Determines the ordering of the section within the TjanTabBar section list.}
      property OnHide          : TNotifyEvent       read FOnHide         write FOnHide; //event
      {Occurs when a section is hidden.}
      property OnShow          : TNotifyEvent       read FOnShow         write FOnShow; //event
      {Occurs when a section is shown.}
   end;

  // -- TjanTabBar ----------------------------------------------
  TjanTabBarEditor = class( TComponentEditor )
    public     // -- Public Declarations --------
    {This is the design time component editor for TjanTabBar allowing you to add new sections and move sections up and down by right-clicking the component.}
      function GetVerbCount : Integer; override;
      {Gets the number of component editor menu items.}
      function GetVerb( Index: Integer ) : string; override;
      {Gets the component editor menu caption.}
      procedure ExecuteVerb( Index: Integer ); override;
      {Executes one of the component editor functions.}
      {}
      {There are three functions defined:}
      {}
      {- Add Section}
      {- Move Section Down}
      {- Move Section Up}
   end;

  // -- TjanTabBar Events ---------------------------------------
  TjanZCChangingEvent = procedure ( Sender: TObject; var AllowChange: Boolean ) of object;
  TjanZoneButtonEvent = procedure ( Sender: TObject; Tab: Integer ) of object;

  // -- TjanTabBar - v 1.0 -------------------------------------
  TjanTabBar = class( TCustomControl )
    private    // -- Private Declarations -------
      FMouseDownTab          : integer;
      FFrameWidth            : Integer;
      FTabHeight             : Integer;
      FOnChange              : TNotifyEvent;
      FOnChanging            : TjanZCChangingEvent;
      FHighLightZone: TjanTabSection;
      FImages: TCustomImageList;
      FAlignment: TAlignment;
      procedure Changer( AZone: TjanTabSection );
      function GetActiveZone : TjanTabSection;
      function GetActiveSection : TjanTabSection;
      function GetZACount : Integer;
      function GetZAAtPos( APoint: TPoint ) : TjanTabSection;
      function GetDisplayRect : TRect;
      function GetTabRect( AIndex: Integer ) : TRect;
      function GetZoneRect : TRect;
      procedure SetActiveZone( AZone: TjanTabSection );
      procedure SetFrameWidth( AValue: Integer );
      procedure SetTabHeight( AValue: Integer );
      procedure ResizeZones;
      procedure SetActiveSection(const Value: TjanTabSection);
      procedure SetImages(Value: TCustomImageList);
      procedure SetAlignment(const Value: TAlignment);
    protected  // -- Protected Declarations -----
      procedure Notification( AComponent: TComponent; Operation: TOperation ); override;
      procedure AdjustClientRect( var Rect: TRect ); override;
      procedure WMGetDlgCode( var Msg: TWMGetDlgCode ); message WM_GETDLGCODE;
      procedure CMDesignHitTest( var Msg: TCMDesignHitTest ); message CM_DESIGNHITTEST;
      procedure CMFontChanged( var Msg: TMessage ); message CM_FONTCHANGED;
      procedure CMMouseLeave(var Message: TMessage); message CM_MouseLeave;
      procedure MouseDown( Button: TMouseButton; Shift: TShiftState; X, Y: Integer ); override;
      procedure MouseUp( Button: TMouseButton; Shift: TShiftState; X, Y: Integer ); override;
      procedure MouseMove(Shift: TShiftState; X, Y: Integer); override;
      procedure KeyDown( var Key: Word; Shift: TShiftState ); override;
      procedure Change;
      procedure ClickAtPos( APoint: TPoint );
      procedure Paint; override;
      function ChildAsZA( AControl: TControl ) : TjanTabSection;
      function ChildIsZA( AControl: TControl ) : Boolean;
    public     // -- Public Declarations --------
      {TjanTabBar is an Outlook style bar with scrollable sections that can be opened and closed by clicking the section tab.}
      constructor Create( AOwner: TComponent ); override;
      {Creates and initializes an instance of TjanTabBar.}
      destructor Destroy; override;
      {Destroys an instance of TjanTabBar.}
      function FindNextTab( AZone: TjanTabSection ) : TjanTabSection;
      {Return the next section.}
      function FindPrevTab( AZone: TjanTabSection ) : TjanTabSection;
      {Return the previous section.}
      property ZACount : Integer read GetZACount;
      {Read only. Gives the number of sections.}
    published  // -- Published Declarations -----
      property Align;
      {Determines how the control aligns within its container (parent control).}
      property Alignment: TAlignment read FAlignment write SetAlignment;
      {Specifies how text is aligned within a control.}
      property ActiveSection: TjanTabSection read GetActiveSection write SetActiveSection;
      {Determines the active section.}
      property Anchors;
      {Specifies how the control is anchored to its parent.}
      property Constraints;
      {Specifies the size constraints for the control.}
      property Ctl3D;
      {Determines whether a control has a three-dimensional (3-D) or two-dimensional look.}
      property FrameWidth      : Integer            read FFrameWidth     write SetFrameWidth;
      {Determines the width of the frame drawn around the control.}
      property Images: TCustomImageList read FImages write SetImages;
      {Lists the images that can appear on section tabs.}
      property TabHeight       : Integer            read FTabHeight      write SetTabHeight;
      {Determines the height of the section tabs.}
      property Parent;
      {Inidicates the parent of the control.}
      property PopupMenu;
      {Identifies the pop-up menu associated with the control.}
      property TabOrder;
      {Indicates the position of the control in its parent's tab order.}
      property TabStop;
      {Determines if the user can tab to a control.}
      property Visible;
      {Determines whether the control appears onscreen.}
      property OnChange        : TNotifyEvent       read FOnChange       write FOnChange; //event
      {Occurs when the control has changed}
      property OnChanging      : TjanZCChangingEvent read FOnChanging     write FOnChanging;  //event
      {Occurs when the control is changing.}
      property OnClick; //event
      {Occurs when the user clicks the control.}
      property OnDblClick; //event
      {Occurs when the user double clicks the control.}
      property OnDragOver; //event
      {Occurs when the user drags an object over a control.}
      property OnEnter;  //event
      {Occurs when a control receives the input focus.}
      property OnExit; //event
      {Occurs when the input focus shifts away from one control to another. }
      property OnKeyDown; //event
      {Occurs when a user presses any key while the control has focus. }
      property OnKeyPress; //event
      {Occurs when a user presses a single character key. }
      property OnKeyUp; //event
      {Occurs when the user releases a key that has been pressed.}
      property OnMouseDown; //event
      {Occurs when the user presses a mouse button with the mouse pointer over a control. }
      property OnMouseMove; //event
      {Occurs when the user moves the mouse pointer while the mouse pointer is over a control.}
      property OnMouseUp;   //event
      {Occurs when the user releases a mouse button that was pressed with the mouse pointer over a component.}
   end;


procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('Jans 2', [TjanTabBar]);
  RegisterNoIcon([TjanTabSection]);
  RegisterClass(TjanTabSection);
  RegisterComponentEditor(TjanTabBar, TjanTabBarEditor);
  RegisterComponentEditor(TjanTabSection, TjanTabBarEditor);
end;

// ------------------------------------------------------------------
// -- TjanTabSection ---------------------------------------------------
// ------------------------------------------------------------------
constructor TjanTabSection.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  Self.BoundsRect := Rect( 10,10,20,20 );
  Self.BorderWidth := 1;
  Self.Visible := True;
  Self.Color := clGray;
  Self.BorderStyle:= bsnone;
  FBackBitmap := TBitmap.Create;
  FBackBitmap.OnChange := BackBitmapChanged;
  FImageIndex:=-1;
  FHighLightColor:=clBlue;
  FTabColor:=clSilver;
end;

destructor TjanTabSection.Destroy;
begin
  inherited Destroy;
  FBackBitmap.free;
end;


// --------------------------------------------------------
// -- Capture Changes to the Display Caption for the Zone
procedure TjanTabSection.CMTextChanged( var Msg: TMessage );
begin
  inherited;
  if Self.ParentIsZC then
    Self.ParentAsZC.Invalidate;
end;

// --------------------------------------------------------
function TjanTabSection.GetTabIndex : Integer;
begin
  Result := Self.TabOrder;
end;

procedure TjanTabSection.SetActive( Value: Boolean );
begin
  if (Value <> FActive) then
  begin
    FActive := Value;
    case Active of
     True : if Assigned(FOnShow) then FOnShow( Self );
       else if Assigned(FOnHide) then FOnHide( Self );
    end;
  end;
end;

procedure TjanTabSection.SetTabIndex( Value: Integer );
begin
  if Self.ParentIsZC then
  if (Value >= 0) and (Value <= (Self.ParentAsZC.ZACount-1)) then
  begin
    Self.TabOrder := Value;
    Self.ParentAsZC.ResizeZones;
    Self.ParentAsZC.Invalidate;
  end;
end;

// --------------------------------------------------------
function TjanTabSection.ParentAsZC : TjanTabBar;
begin
  Result := TjanTabBar( Self.Parent );
end;

function TjanTabSection.ParentIsZC : Boolean;
begin
  Result := (Self.Parent <> nil) and (Self.Parent is TjanTabBar);
end;


// ------------------------------------------------------------------
// -- TjanTabBar Contextual Menu Editor Controller --------------
// ------------------------------------------------------------------
function TjanTabBarEditor.GetVerbCount: Integer;
begin
  Result := 3;
end;

function TjanTabBarEditor.GetVerb( Index: Integer ) : string;
begin
  case Index of
   0 : Result := 'New &Area';
   1 : Result := 'Move &Down';
   2 : Result := 'Move &Up';
  end;
end;

procedure TjanTabBarEditor.ExecuteVerb( Index: Integer );
var
  ovDesigner                 : IFormDesigner;
  ovZoneControl              : TjanTabBar;
  ovNewZone, ovMoveZone      : TjanTabSection;
  i: integer;
begin
  ovDesigner := Self.Designer;
  case Index of
   0 : begin   // -- New Area -------------------
         if (Component is TjanTabBar) then
              ovZoneControl := (Component as TjanTabBar)
         else ovZoneControl := (TjanTabSection(Component).Parent as TjanTabBar);
         ovNewZone := TjanTabSection.Create( ovDesigner.Form );
         ovNewZone.Name := ovDesigner.UniqueName( TjanTabSection.ClassName );
         ovNewZone.Caption := ovNewZone.Name;
         ovNewZone.Parent := ovZoneControl;
         ovZoneControl.ActiveSection:=ovNewZone;
         ovDesigner.SelectComponent( ovNewZone );
         ovNewZone.ParentAsZC.Invalidate;
       end;
   1 : begin   // move tab down
         if (Component is TjanTabSection) then
         begin
            ovZoneControl := (TjanTabSection(Component).Parent as TjanTabBar);
            ovNewZone:= TjanTabSection(Component);
            if ovNewZone.TabIndex< ovZoneControl.ZACount-1 then
            begin
              ovMoveZone:=ovZoneControl.FindNextTab(ovNewZone);
              i:=ovNewZone.TabIndex;
              ovNewZone.TabIndex:=ovMoveZone.tabindex;
              ovMoveZone.tabindex:= i;
              ovZoneControl.Invalidate;
            end;
         end;
       end;
   2 : begin   // move tab up
         if (Component is TjanTabSection) then
         begin
            ovZoneControl := (TjanTabSection(Component).Parent as TjanTabBar);
            ovNewZone:= TjanTabSection(Component);
            if ovNewZone.TabIndex>0 then
            begin
              ovMoveZone:=ovZoneControl.FindPrevTab(ovNewZone);
              i:=ovNewZone.TabIndex;
              ovNewZone.TabIndex:=ovMoveZone.tabindex;
              ovMoveZone.tabindex:= i;
              ovZoneControl.Invalidate;
            end;
         end;
       end;
  end;
  Designer.Modified;
end;


// ------------------------------------------------------------------
// -- TjanTabBar ------------------------------------------------
// ------------------------------------------------------------------
constructor TjanTabBar.Create( AOwner: TComponent );
begin
  inherited Create( AOwner );
  FFrameWidth := 1;
  Self.BoundsRect := Rect( 10,10,140,110 );
  Self.Align := alLeft;
  Self.ControlStyle := [ csCaptureMouse, csClickEvents, csOpaque,
                         csDoubleClicks ];
  Self.Color := clGray;
  Self.TabStop := True;
  FTabHeight := 20;
  FMouseDownTab:=-1;
end;

destructor TjanTabBar.Destroy;
begin
  inherited Destroy;
end;

// --------------------------------------------------------
procedure TjanTabBar.Notification( AComponent: TComponent; Operation: TOperation );
begin
  inherited Notification( AComponent, Operation );
  if (AComponent is TjanTabSection) then
  case Operation of
   opInsert : SetActiveZone( TjanTabSection(AComponent) );
   opRemove :
     begin
      if Acomponent=fImages then
        SetImages(nil)
      else if (Self.ControlCount <= 0) then
        SetActiveZone( nil )
      else if ChildIsZA( Self.Controls[0] ) then
        SetActiveZone( ChildAsZA( Self.Controls[0] ) );
     end;
  end;
end;

procedure TjanTabBar.AdjustClientRect( var Rect: TRect );
begin
  Rect := GetDisplayRect;
  InflateRect( Rect, -FFrameWidth, -FFrameWidth );
  InflateRect( Rect, 0, -100 );
  inherited AdjustClientRect( Rect );
end;

// -- Sets the component to TRAP the requested element types
procedure TjanTabBar.WMGetDlgCode( var Msg: TWMGetDlgCode );
begin   // _WANTARROWS,  _WANTTAB,  _WANTCHARS
  Msg.Result := (DLGC_WANTARROWS + DLGC_WANTCHARS);
end;

// -- Captures Mouse movement and Clicking during Design-Time
procedure TjanTabBar.CMDesignHitTest( var Msg: TCMDesignHitTest );
const
  ovControlClick = 1;
begin
  inherited;
  if (Msg.Keys = ovControlClick) then
    ClickAtPos( Point( Msg.Pos.x, Msg.Pos.y ) );
end;

// -- Captures Changes to the Display Font
procedure TjanTabBar.CMFontChanged( var Msg: TMessage );
begin
  inherited;
  Invalidate;
end;

procedure TjanTabBar.MouseDown( Button: TMouseButton; Shift: TShiftState;
                                    X, Y: Integer );
var
  ovZone                     : TjanTabSection;
begin
  inherited;
  ovZone := GetZAAtPos( point(x,y));
  if (ovZone <> nil) then
    FMouseDownTab:=ovZone.TabIndex
  else
    FMouseDownTab:=-1;
  if (not (csDesigning in Self.ComponentState)) and
     (Self.CanFocus or (GetParentForm(Self) = nil)) then
  begin
    Self.SetFocus;
    if (not Self.Focused) then Exit;
  end;
  if FMouseDownTab<>-1 then
    invalidate;
//  ClickAtPos( Point( X, Y ) );
end;

procedure TjanTabBar.KeyDown( var Key: Word; Shift: TShiftState );
begin
  inherited KeyDown(Key, Shift);
  if (Key in [VK_UP,VK_DOWN]) then
  begin
    case Key of
      VK_UP : changer( FindPrevTab( ActiveSection ));
      VK_DOWN : changer(FindNextTab( ActiveSection ));
    end;
  end;
end;

// --------------------------------------------------------
procedure TjanTabBar.Change;
begin
  inherited;
  Changer( Self.ActiveSection );
end;

procedure TjanTabBar.Changer( AZone: TjanTabSection );
var
  ovForm                     : TCustomForm;
  ovChangeOK, ovInDesign     : Boolean;
begin
  ovForm := GetParentForm( Self );     //-- D/T Property/Component Editors Link
  ovChangeOK := True;                  //-- State Processing Switch Definition
  ovInDesign := (csDesigning in Self.ComponentState);
  if (not ovInDesign) and Assigned(FOnChanging) then
    FOnChanging( Self, ovChangeOK );   //-- Check validity of Zone Change
  if ovChangeOK then
  begin
    if (AZone <> nil) then SetActiveZone( AZone )      //-- Update Active Zone
                      else SetActiveZone( nil );
    if ovInDesign and (ovForm <> nil) then             //-- Update D/T Editors
      if (ovForm.Designer <> nil) then ovForm.Designer.Modified;
    if AZone.Active and (AZone = Self.ActiveSection) then //-- Display Active Zone
         AZone.BringToFront
    else AZone.SendToBack;
    if (not ovInDesign) and Assigned(FOnChange) then
      FOnChange( Self );               //-- Notify User of Zone Change
  end;
end;

procedure TjanTabBar.ClickAtPos( APoint: TPoint );
var
  ovZone                     : TjanTabSection;
begin
  ovZone := GetZAAtPos( APoint );
  if (ovZone <> nil) then
  if (ActiveSection = nil) or (ActiveSection <> ovZone) then
    Changer( ovZone )
  else
    invalidate;  
end;

// --------------------------------------------------------
function TjanTabBar.GetActiveZone : TjanTabSection;
var
  ovZone                     : TjanTabSection;
  ovLoop                     : Integer;
begin
  Result := nil;
  for ovLoop := 0 to (Self.ControlCount-1) do
  if ChildIsZA( Self.Controls[ovLoop] ) then
  begin
    ovZone := ChildAsZA( Self.Controls[ovLoop] );
    if ovZone.Active then Result := ovZone;
  end;
end;

function TjanTabBar.GetZACount : Integer;
var
  ovLoop                     : Integer;
begin
  Result := 0;
  for ovLoop := 0 to (Self.ControlCount-1) do
  if ChildIsZA( Self.Controls[ovLoop] ) then
    Inc( Result );
end;

function TjanTabBar.GetZAAtPos( APoint: TPoint ) : TjanTabSection;
var
  ovZone                     : TjanTabSection;
  ovLoop                     : Integer;
begin
  Result := nil;
  for ovLoop := 0 to (Self.ControlCount-1) do
  if ChildIsZA( Self.Controls[ovLoop] ) then
  begin
    ovZone := ChildAsZA( Self.Controls[ovLoop] );
    if PtInRect(GetTabRect(ovZone.TabIndex), APoint) then Result := ovZone;
  end;
end;

function TjanTabBar.GetDisplayRect : TRect;
begin
  Result := Self.ClientRect;
end;

function TjanTabBar.GetTabRect( AIndex: Integer ) : TRect;
begin
  if (AIndex < 0) or (AIndex > (ZACount-1)) then
       Result := Rect( 0,0,0,0 )
  else begin
         Result := Rect( 2, 0, Width-2, 1 );
         if (AIndex <= ActiveSection.TabIndex) then
              Result.Top := ((AIndex * TabHeight) + 1)
         else Result.Top := (Height - ((ZACount - AIndex) * TabHeight) - 1);
         Result.Bottom := (Result.Top + TabHeight - 1);
       end;
end;

function TjanTabBar.GetZoneRect : TRect;
var
  ovActiveIndex              : Integer;
begin
  if (ActiveSection = nil) then
       ovActiveIndex := 0
  else ovActiveIndex := ActiveSection.TabIndex;
  if (ovActiveIndex < 0) or (ovActiveIndex > (ZACount-1)) then
       Result := Rect( 0,0,0,0 )
  else begin
         Inc( ovActiveIndex );
         Result := Rect( 1, 0, (Width - 2), 0 );
         Result.Top := (ovActiveIndex * TabHeight);
         Result.Bottom := (Height-(Max(0,(ZACount-ovActiveIndex)) * TabHeight)-1);
       end;
end;

procedure TjanTabBar.SetActiveZone( AZone: TjanTabSection );
var
  ovZone                     : TjanTabSection;
  ovLoop                     : Integer;
begin
  for ovLoop := (Self.ControlCount-1) downto 0 do
  if ChildIsZA( Self.Controls[ovLoop] ) then
  begin
    ovZone := ChildAsZA( Self.Controls[ovLoop] );
    ovZone.Active := (AZone = ovZone);
  end;
end;

procedure TjanTabBar.SetFrameWidth( AValue: Integer );
begin
  if (AValue <> FFrameWidth) and (AValue in [0..10]) then
  begin
    FFrameWidth := AValue;
    Self.Invalidate;
  end;
end;


procedure TjanTabBar.SetTabHeight( AValue: Integer );
begin
  if (AValue <> FTabHeight) and (AValue in [0..50]) then
  begin
    FTabHeight := AValue;
    Self.Invalidate;
  end;
end;

// --------------------------------------------------------
procedure TjanTabBar.Paint;
var
  ovZone                     : TjanTabSection;
  ovRect, Rcap               : TRect;
  ovSize                     : TSize;
  ovLoop                     : Integer;
  ovCaption                  : string;
  glyphw:integer; // glyphwidth

  procedure DrawGlyph( bmZone:TjanTabSection;bmTab:TRect);
  begin
    Fimages.Draw(self.Canvas,bmtab.Left+2,bmtab.top+2,bmZone.imageindex);
  end;


  procedure DrawBackBitmap( bmZone:TjanTabSection;bmTab:TRect);
  var
    ix, iy,ww,hh: Integer;
    BmpWidth, BmpHeight: Integer;
    x0,y0:integer;
    TabW,TabH:integer;
    hCanvas, BmpCanvas: THandle;
    bm: Tbitmap;
    begin
      bm := bmzone.FBackBitmap;
      tabW:=bmTab.Right-bmTab.Left;
      tabH:=bmTab.Bottom-bmTab.Top;
      BmpWidth := bm.Width;
      BmpHeight := bm.Height;
      BmpCanvas := bm.Canvas.Handle;
      hCanvas := THandle(self.canvas.handle);
      if TabW>BmpWidth then ww:=BmpWidth else ww:=TabW;
      if TabH>bmpHeight then hh:=BmpHeight else hh:=TabH;
      y0:=bmTab.top;
      repeat
        x0:=bmTab.left;
        if (y0+BmpHeight)<(bmTab.bottom-1) then
          hh:=BmpHeight
        else
          hh:=bmtab.bottom-1-y0;
        repeat
          if (x0+BmpWidth)<(bmTab.right-1) then
            ww:=BmpWidth
          else
            ww:=bmtab.right-1-x0;
          BitBlt(hCanvas, x0, y0, ww, hh, BmpCanvas,0, 0, SRCCOPY);
          x0:=x0+ww;
        until x0>=bmTab.right-1;
        y0:=y0+hh;
      until y0>=bmTab.bottom-1;
  end;

begin
  inherited;
  ResizeZones;
  ovRect := Self.ClientRect;
  with Self.Canvas do
  begin
    Pen.Style := psSolid;
    Pen.Width := 1;
    Pen.Color :=  clBlack;
    Brush.color:=clsilver;
    if (Self.ControlCount = 0) then FillRect( ovRect );
  end;
  Frame3D( Self.Canvas, ovRect, clBlack, clBtnHighlight, FFrameWidth );
  if assigned(FImages) then
    GlyphW:=Fimages.Width
  else
    GlyphW:=0;
  with Self.Canvas do
    for ovLoop := (Self.ControlCount-1) downto 0 do
    if ChildIsZA( Self.Controls[ovLoop] ) then
    begin
      ovZone := ChildAsZA( Self.Controls[ovLoop] );
      ovRect := GetTabRect( ovZone.TabIndex );
      Font := ovzone.Font;
      brush.color:=ovZone.FTabColor;
      if assigned(ovzone.FBackBitmap) and (ovzone.FBackBitmap.Height <> 0) and (ovZone.FBackBitmap.Width <> 0) then
        DrawBackBitmap(ovZone,ovRect)
      else
        FillRect( ovRect );
      if assigned(FImages)and (ovZone.FImageIndex>=0) then
        DrawGlyph(ovzone,ovRect);
      ovCaption := ovZone.Caption;
      if ovZone=FHighLightZone then
        font.color:=ovzone.HighLightColor
      else
        font.color:=ovZone.font.color;
//      ovSize := TextExtent( ovCaption );
      brush.style:=bsclear;
      Rcap:=ovRect;
      inflateRect(Rcap,-2,-1);
      Rcap.Left :=Rcap.left+GlyphW+2;
      case Alignment of
      taleftjustify:
        DrawText(self.canvas.handle,@ovCaption[1],-1,Rcap,DT_LEFT or DT_SINGLELINE or DT_END_ELLIPSIS);
      tacenter:
        DrawText(self.canvas.handle,@ovCaption[1],-1,Rcap,DT_CENTER or DT_SINGLELINE or DT_END_ELLIPSIS);
      taRightJustify:
        DrawText(self.canvas.handle,@ovCaption[1],-1,Rcap,DT_RIGHT or DT_SINGLELINE or DT_END_ELLIPSIS);
      end;
//      TextOut( GlyphW+((Self.Width-3-GlyphW-ovSize.cx) div 2),
//               ovRect.Top+((Self.TabHeight-ovSize.cy) div 2), ovCaption );
      if ovZone.TabIndex <>FMouseDownTab then
        Frame3D( Self.Canvas, ovRect, clBtnHighlight, clBlack, 1 )
      else
        Frame3D( Self.Canvas, ovRect, clblack, clBtnHighlight, 1 );
    end;
end;

// --------------------------------------------------------
procedure TjanTabBar.ResizeZones;
var
  ovRect                     : TRect;
  ovLoop                     : Integer;
begin
  ovRect := GetZoneRect;
  for ovLoop := 0 to (Self.ControlCount-1) do
  if ChildIsZA( Self.Controls[ovLoop] ) then
    ChildAsZA( Self.Controls[ovLoop] ).BoundsRect := ovRect;
end;

// --------------------------------------------------------
function TjanTabBar.FindNextTab( AZone: TjanTabSection ) : TjanTabSection;
var
  ovLoop, ovCount, ovFindTab : Integer;
begin
  ovCount := Self.ZACount;
  if (ovCount < 2) or (AZone = nil) then
       Result := Self.ActiveSection
  else begin
         Result := nil;
         ovFindTab := AZone.TabIndex + 1;
         if (ovFindTab >= ovCount) then ovFindTab := 0;
         for ovLoop := 0 to (Self.ControlCount-1) do
         if ChildIsZA( Self.Controls[ovLoop] ) then
         if (ChildAsZA( Self.Controls[ovLoop] ).TabIndex = ovFindTab) then
           Result := ChildAsZA( Self.Controls[ovLoop] );
       end;
end;

function TjanTabBar.FindPrevTab( AZone: TjanTabSection ) : TjanTabSection;
var
  ovLoop, ovCount, ovFindTab : Integer;
begin
  ovCount := Self.ZACount;
  if (ovCount < 2) or (AZone = nil) then
       Result := Self.ActiveSection
  else begin
         Result := nil;
         ovFindTab := AZone.TabIndex - 1;
         if (ovFindTab < 0) then ovFindTab := ovCount - 1;
         for ovLoop := 0 to (Self.ControlCount-1) do
         if ChildIsZA( Self.Controls[ovLoop] ) then
         if (ChildAsZA( Self.Controls[ovLoop] ).TabIndex = ovFindTab) then
           Result := ChildAsZA( Self.Controls[ovLoop] );
       end;
end;

// --------------------------------------------------------
function TjanTabBar.ChildAsZA( AControl: TControl ) : TjanTabSection;
begin
  Result := TjanTabSection( AControl );
end;

function TjanTabBar.ChildIsZA( AControl: TControl ) : Boolean;
begin
  Result := (AControl <> nil) and (AControl is TjanTabSection);
end;

// ------------------------------------------------------------------




procedure TjanTabBar.MouseMove(Shift: TShiftState; X, Y: Integer);
var
  ovZone                     : TjanTabSection;
begin
  if (csDesigning in ComponentState) and (FMouseDownTab<>-1) then
  begin
    self.BeginDrag (true);
    exit;
  end;
  ovZone := GetZAAtPos(point(x,y));
  if FHighLightZone<>ovZone then
  begin
    FHighLightZone:=ovzone;
    invalidate;
  end;
end;

procedure TjanTabBar.CMMouseLeave(var Message: TMessage);
begin
   FHighLightZone:=nil;
   invalidate;
end;

procedure TjanTabBar.MouseUp(Button: TMouseButton; Shift: TShiftState;
  X, Y: Integer);
begin
  inherited;
  FMouseDownTab:=-1;
  ClickAtPos( Point( X, Y ) );
end;

procedure TjanTabSection.SetBackBitmap(const Value: TBitmap);
begin
  FBackBitmap.assign(Value);
end;

procedure TjanTabSection.BackBitmapChanged(Sender: TObject);
begin
   self.ParentAsZC.Invalidate;
end;

procedure TjanTabSection.CMFontChanged(var Msg: TMessage);
begin
  if Self.ParentIsZC then
    Self.ParentAsZC.Invalidate;
end;

procedure TjanTabSection.SetHighLightColor(const Value: TColor);
begin
  FHighLightColor := Value;
end;

procedure TjanTabSection.SetTabColor(const Value: TColor);
begin
  FTabColor := Value;
  if Self.ParentIsZC then
    Self.ParentAsZC.Invalidate;
end;

procedure TjanTabBar.SetActiveSection(const Value: TjanTabSection);
begin
   changer(Value);
end;

function TjanTabBar.GetActiveSection: TjanTabSection;
begin
   result:=GetActiveZone;
end;

procedure TjanTabBar.SetImages(Value: TCustomImageList);
begin
  FImages := Value;
  invalidate;
end;




procedure TjanTabSection.SetImageIndex(Value: TImageIndex);
begin
  if FImageIndex <> Value then
  begin
    FImageIndex := Value;
    if assigned(self.ParentAsZC.FImages) then
      self.ParentAsZC.Invalidate;
  end;
end;


procedure TjanTabBar.SetAlignment(const Value: TAlignment);
begin
  if value<>Falignment then
  begin
    FAlignment := Value;
    invalidate;
  end;
end;


end.
